/*
 * Real time packet sending client
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "common.h"
#include "recv_packet.h"
#include "send_packet.h"
#include "tracer.h"

// Structs

enum TSNTask { SEND_PACKET_TASK, RTT_TASK };

typedef struct rtt_stat {
  int min_rtt;
  int avg_rtt;
  int max_rtt;

} rtt_stat_t;

typedef struct thread_param {
  uint64_t interval;
  unsigned int max_cycles;
  int priority;
  int etf_offset;
  int64_t send_ts_delay;
  int affinity_cpu;
  uint64_t start_ts;

  int enable_send_ts;
  int enable_etf_tracing;
  int enable_threshold_tracing;
  int threshold;

} thread_param_t;

typedef struct main_param {
  int refresh_rate;
  int verbose;
  int enable_tracing;
  int enable_graph;
} main_param_t;

// Static functions

static void process_options(int argc, char *argv[]);
static void do_tsn_task(char *data, uint64_t next_txtime);
static void print_histograms();
static void sighand(int sig_num);

// Static variables

static uint64_t kernel_latency_hist[MAX_KERNEL_LATENCY];
static uint64_t rtt_hist[MAX_RTT];
static uint64_t nb_cycles;
static main_param_t main_params;
static thread_param_t thread_params;
static egress_param_t egress_params;
static ingress_param_t ingress_params;
static rtt_stat_t rtt_stats = {.min_rtt = INT_MAX};
static egress_stat_t egress_stats = {.min_kernel_latency = INT_MAX,
                                     .min_interval = INT_MAX};
static ingress_stat_t ingress_stats = {.min_kernel_latency = INT_MAX,
                                       .min_interval = INT_MAX};
static int enable_histograms;
static int enable_etf;
static int enable_timestamps;
static enum TSNTask tsn_task;
static struct timespec measures_start;
static struct timespec measures_end;
static char send_data[MAX_BUFFER_SIZE];

static void help(char *argv[]) {
  printf(
      "Usage: %s [-a CPU -p PRIO -i USEC -r USEC -l N] [-d BUF_LEN | -c "
      "DELAY -s NS] [-b -e OFFSET -q PK_PRIO -gtvT] IF IP\n\n"
      "       -h                 Show help\n"
      "       -f IF              Set the network interface to be used\n"
      "       -a CPU             Pin the real time thread to CPU\n"
      "       -p PRIO            RT thread priority\n"
      "       -i USEC            RT thread wake-up interval (default: 10ms)\n"
      "       -r USEC            non-RT main thread refresh interval\n"
      "       -l N_CYCLES        RT thread cycles amount (default: infinite)\n"
      "       -d BUF_LEN         TX buffer len\n"
      "       -c USEC            Send timestamp with the specified delay\n"
      "       -s NS              Common start time reference\n"
      "       -b                 Measure RTT\n"
      "       -e USEC            txtime userspace offset (for ETF qdisc)\n"
      "       -q PK_PRIO         Packet priority \n"
      "       -g                 Print histograms to sdtout on exit\n"
      "       -t                 Enable TX timestamps\n"
      "       -v                 Verbose\n"
      "       -T                 Enable tracing until deadline is missed\n"
      "       -S                 Enable tracing until threshold is reached\n"
      "\n",
      argv[0]);
}

/*
 * Real-time thread:
 *   - Sends packets at a regular intervall
 */
static void *packet_sending_thread(void *p) {
  (void)p;
  struct timespec next, current, previous;
  uint64_t next_txtime = 0, tx_data = 0;
  int ret;
  cpu_set_t mask;

  // Set thread CPU affinity
  if (thread_params.affinity_cpu != -1) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  clock_gettime(CLOCK_MONOTONIC, &measures_start);
  clock_gettime(CLOCK_REALTIME, &next);

  if (thread_params.start_ts) {
    if (thread_params.start_ts < ts_to_uint(next)) {
      fprintf(stderr, "start timestamp is in the past, aborting...\n");
      exit(EXIT_FAILURE);
    }
    if (thread_params.start_ts > (ts_to_uint(next) + UINT64_C(3600000000000))) {
      fprintf(stderr, "start timestamp is too high, aborting...\n");
      exit(EXIT_FAILURE);
    }
    next = uint_to_ts(thread_params.start_ts);
  }
  add_ns(&next, thread_params.interval);

  // Set txtime and TX data timestamp if required
  if (enable_etf || thread_params.enable_send_ts) {
    if (enable_etf) {
      next_txtime = next.tv_sec * NSEC_PER_SEC + next.tv_nsec;
      next_txtime += thread_params.etf_offset;
    }

    if (thread_params.enable_send_ts) {
      tx_data = next.tv_sec * NSEC_PER_SEC + next.tv_nsec;
      tx_data += thread_params.send_ts_delay;
    }
  }

  printf("Will start at %" PRIu64 "\n", ts_to_uint(next));

  // Packet sending loop
  for (nb_cycles = 0;; nb_cycles++) {
    if (thread_params.max_cycles &&
        nb_cycles >= ((unsigned int)thread_params.max_cycles))
      break;

    clock_gettime(CLOCK_REALTIME, &current);

    // Sanity check
    if(current.tv_sec > next.tv_sec)
      goto invalid_ts;
    if(current.tv_sec == next.tv_sec && current.tv_nsec > next.tv_nsec)
      goto invalid_ts;

    ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &next, NULL);
    if (ret) {
      fprintf(stderr, "clock_nanosleep returned error: %d, aborting...\n", ret);
      exit(EXIT_FAILURE);
    }

    // TX data
    encode(thread_params.enable_send_ts ? tx_data : 0xdeadbeef, send_data);

    // Get timestamp before TSN task for stats
    clock_gettime(CLOCK_REALTIME, &current);
    do_tsn_task(send_data, next_txtime);
    add_ns(&next, thread_params.interval);

    // Update txtime and TX data timestamp if required
    if (enable_etf) next_txtime += thread_params.interval;
    if (thread_params.enable_send_ts) {
      tx_data = next.tv_sec * NSEC_PER_SEC + next.tv_nsec;
      tx_data += thread_params.send_ts_delay;
    }

    // Update statistics
    if (nb_cycles) {
      int interval_us = calcdiff_ns(current, previous) / 1000;

      egress_stats.min_interval = _min_(interval_us, egress_stats.min_interval);
      egress_stats.max_interval = _max_(interval_us, egress_stats.max_interval);
      egress_stats.avg_interval =
          (egress_stats.avg_interval * nb_cycles + interval_us) /
          (nb_cycles + 1);

      if (thread_params.enable_threshold_tracing) {
        int jitter = (egress_stats.max_interval - egress_stats.min_interval);
        if (jitter > thread_params.threshold) {
          tracemark("Threshold reached\n");
          tracing(0);
          printf("Threshold reached: %dus\n", jitter);
          printf("Exiting packet-exchange...\n");
          exit(EXIT_SUCCESS);
        }
      }
    }

    previous = current;
  }

  pthread_exit(NULL);

invalid_ts:
  fprintf(stderr, "Invalid timestamp         : %" PRIu64 "\n", ts_to_uint(next));
  fprintf(stderr, "Current time is           : %" PRIu64 "\n", ts_to_uint(current));
  fprintf(stderr, "Difference with timestamp : %" PRIi64 "\n\n\n", ((int64_t)ts_to_uint(next)) - ((int64_t)ts_to_uint(current)));
  exit(EXIT_FAILURE);
}

/*
 * Main thread:
 *  - Has non-real time priority
 *  - Handles the IO and creates the real time thread
 */
int main(int argc, char *argv[]) {
  pthread_t thread;
  struct sched_param param;
  pthread_attr_t attr;

  // Default configuration values
  thread_params.interval = 100000 * 1000;
  thread_params.max_cycles = 0;
  thread_params.priority = 98;
  thread_params.send_ts_delay = 0;
  thread_params.affinity_cpu = -1;
  thread_params.start_ts = 0;
  thread_params.enable_etf_tracing = 0;
  thread_params.enable_threshold_tracing = 0;
  thread_params.enable_send_ts = 0;
  main_params.refresh_rate = 50000;
  main_params.verbose = 0;
  main_params.enable_tracing = 0;
  main_params.enable_graph = 0;
  egress_params.packet_priority = 3;
  egress_params.tx_buffer_len = 1024;
  enable_etf = 0;
  enable_timestamps = 0;
  enable_histograms = 0;
  tsn_task = SEND_PACKET_TASK;

  // Lock all current and future pages from preventing of being paged to
  // swap
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    fprintf(stderr, "mlockall failed");
    exit(EXIT_FAILURE);
  }

  // Process bash options
  process_options(argc, argv);

  set_latency_target();

  egress_params.use_etf = enable_etf;
  egress_params.use_timestamps = enable_timestamps;

  if (enable_histograms) {
    // Init histograms
    memset(kernel_latency_hist, 0, sizeof(kernel_latency_hist));
    memset(rtt_hist, 0, sizeof(rtt_hist));
  }

  if (main_params.enable_tracing) {
    // Open tracing file descriptors
    open_fds();
  }

  // Catch breaks with sighand to print the histograms
  init_signals(sighand);

  // Initialize the UDP packet sending socket
  init_udp_send(&egress_params, &egress_stats, enable_histograms,
                thread_params.enable_etf_tracing, kernel_latency_hist);

  // Initialize the UDP packet receiving socket if RTT is measured
  if (tsn_task == RTT_TASK)
    init_udp_recv(&ingress_params, &ingress_stats, enable_histograms,
                  kernel_latency_hist);

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&attr)) {
    fprintf(stderr, "init pthread attributes failed\n");
    exit(EXIT_FAILURE);
  }
  // Set a specific stack size
  if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN)) {
    fprintf(stderr, "pthread setstacksize failed\n");
    exit(EXIT_FAILURE);
  }
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
    fprintf(stderr, "pthread setschedpolicy failed\n");
    exit(EXIT_FAILURE);
  }
  param.sched_priority = thread_params.priority;
  if (pthread_attr_setschedparam(&attr, &param)) {
    fprintf(stderr, "pthread setschedparam failed\n");
    exit(EXIT_FAILURE);
  }
  /* Use scheduling parameters of attr */
  if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
    fprintf(stderr, "pthread setinheritsched failed\n");
    exit(EXIT_FAILURE);
  }

  // Create the real time thread
  if (pthread_create(&thread, &attr, packet_sending_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create packet sending thread");

  // Verbose loop
  for (;;) {
    usleep(main_params.refresh_rate);

    if (main_params.verbose) {
      // RTT stats printing
      if (tsn_task == RTT_TASK) {
        // N_CYCLES, RTT min / avg / max
        printf("%9" PRIu64 ": RTT: %4d %4d %4d\n", nb_cycles, rtt_stats.min_rtt,
               rtt_stats.avg_rtt, rtt_stats.max_rtt);
        printf("\033[%dA", 1);

        // Packet send stats printing
      } else {
        // N_CYCLES, error queue, measured send interval min / avg / max
        printf("%9" PRIu64 ": [%4d, %4d], I (10us): %3d %3d %3d", nb_cycles,
               (int)egress_stats.invalid_parameter,
               (int)egress_stats.missed_deadline,
               egress_stats.min_interval / 10, egress_stats.avg_interval / 10,
               egress_stats.max_interval / 10);

        // SOF_TX timestamps
        if (enable_timestamps)
          printf(", K: %4d %4d %4d [%4d]\n", egress_stats.min_kernel_latency,
                 egress_stats.avg_kernel_latency,
                 egress_stats.max_kernel_latency,
                 (int)egress_stats.high_kernel_latency);
        else
          printf("\n");

        printf("\033[%dA", 1);
      }
    }

    if (thread_params.max_cycles)
      if (thread_params.max_cycles == nb_cycles) break;
  }

  if (enable_histograms) print_histograms();

  exit(EXIT_SUCCESS);
}

/* Critical TSN task
 */
static void do_tsn_task(char *data, uint64_t next_txtime) {
  struct timespec t1, t2;
  int rtt_us;

  // One way packet sending
  if (tsn_task == SEND_PACKET_TASK) {
    send_udp_packet(data, next_txtime);

    // Round Trip Time measurement
  } else if (tsn_task == RTT_TASK) {
    clock_gettime(CLOCK_MONOTONIC, &t1);
    send_udp_packet(data, next_txtime);
    recv_udp_packet();
    clock_gettime(CLOCK_MONOTONIC, &t2);

    rtt_us = calcdiff_ns(t2, t1) / 1000;

    rtt_stats.min_rtt = _min_(rtt_us, rtt_stats.min_rtt);
    rtt_stats.max_rtt = _max_(rtt_us, rtt_stats.max_rtt);
    rtt_stats.avg_rtt =
        (((uint64_t)rtt_stats.avg_rtt) * (nb_cycles - 1) + rtt_us) / nb_cycles;

    if (rtt_us > MAX_RTT)
      fprintf(stderr, "RTT value higher than MAX_RTT : %d ( > %d)\n", rtt_us,
              MAX_RTT);
    else
      rtt_hist[rtt_us]++;
  }
}

/* Print histogram
 */
static void print_histograms() {
  uint64_t duration;
  int duration_hour, duration_minutes;
  int max_hist_val;
  uint64_t *histogram;

  clock_gettime(CLOCK_MONOTONIC, &measures_end);

  duration = calcdiff_ns(measures_end, measures_start);
  duration_hour = duration / NSEC_PER_SEC / 3600;
  duration_minutes = duration / NSEC_PER_SEC / 60 - duration_hour * 60;

  if (tsn_task == SEND_PACKET_TASK) {
    histogram = kernel_latency_hist;
    max_hist_val = histogram_max(histogram, MAX_KERNEL_LATENCY - 1);
  } else {
    histogram = rtt_hist;
    max_hist_val = histogram_max(histogram, MAX_RTT - 1);
  }

  printf("# Packet TX timestamps histogram\n");

  for (int j = 0; j <= max_hist_val; j++)
    printf("%06d %015" PRIi64 "\n", j, histogram[j]);

  printf(
      "# Duration: %dh%d\n"
      "# High kernel latencies: %" PRIu64
      "\n"
      "# Invalid parameters: %" PRIu64
      "\n"
      "# Missed deadlines: %" PRIu64 "\n",
      duration_hour, duration_minutes, egress_stats.high_kernel_latency,
      egress_stats.invalid_parameter, egress_stats.missed_deadline);
}

static void sighand(int sig_num) {
  (void)sig_num;
  if (enable_histograms) print_histograms();
  exit(EXIT_SUCCESS);
}

/* Process bash options
 */
static void process_options(int argc, char *argv[]) {
  for (;;) {
    int c = getopt(argc, argv, "a:bc:d:e:ghi:l:p:q:r:s:tvTS:");

    if (c == -1) break;

    switch (c) {
      case 'a':
        thread_params.affinity_cpu = atoi(optarg);
        break;
      case 'b':
        tsn_task = RTT_TASK;
        break;
      case 'c':
        thread_params.send_ts_delay = strtoll(optarg, NULL, 10) * 1000;
        thread_params.enable_send_ts = 1;
        break;
      case 'd':
        egress_params.tx_buffer_len = atoi(optarg);
        if (egress_params.tx_buffer_len < 1) {
          fprintf(stderr, "BUF_LEN should be greater than 1\n");
          exit(EXIT_FAILURE);
        }
        break;
      case 'e':
        enable_etf = 1;
        thread_params.etf_offset = atoi(optarg) * 1000;
        break;
      case 'g':
        enable_histograms = 1;
        break;
      case 'h':
        help(argv);
        exit(EXIT_SUCCESS);
        break;
      case 'i':
        thread_params.interval = strtoull(optarg, NULL, 10) * 1000;
        break;
      case 'l':
        thread_params.max_cycles = atoi(optarg);
        break;
      case 'p':
        thread_params.priority = atoi(optarg);
        break;
      case 'q':
        egress_params.packet_priority = atoi(optarg);
        break;
      case 's':
        thread_params.start_ts = strtoull(optarg, NULL, 10);
        break;
      case 'r':
        main_params.refresh_rate = atoi(optarg);
        break;
      case 't':
        enable_timestamps = 1;
        break;
      case 'v':
        main_params.verbose = 1;
        break;
      case 'T':
        main_params.enable_tracing = 1;
        thread_params.enable_etf_tracing = 1;
        break;
      case 'S':
        main_params.enable_tracing = 1;
        thread_params.enable_threshold_tracing = 1;
        thread_params.threshold = atoi(optarg);
        break;
    }
  }

  if (argc != optind + 2) {
    if (argc < optind + 2)
      fprintf(stderr, "You need to specifiy an interface and IP address\n");
    else
      fprintf(stderr, "Too many arguments\n");
    help(argv);
    exit(EXIT_FAILURE);
  }

  strcpy(egress_params.network_if, argv[optind]);
  strcpy(egress_params.server_ip, argv[optind + 1]);
}
