/*
 * Software PPS
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <limits.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <time.h>
#include <unistd.h>

#include "common.h"
#include "gpio.h"
#include "pulse.h"

// Structs

typedef struct thread_param {
  uint64_t interval;
  unsigned int max_cycles;
  int priority;
  int affinity_cpu;
  int poll;
  int poll_margin;
  uint64_t start_ts;

} thread_param_t;

typedef struct main_param {
  int refresh_rate;
  int verbose;

  int interval_input;
  int transition_time;
  uint64_t target_interval;

} main_param_t;

// Static functions

static void process_options(int argc, char *argv[]);
static uint64_t nb_cycles;

// Static variables

static main_param_t main_params;
static thread_param_t thread_params;
static int use_gpio;

static void help(char *argv[]) {
  printf(
      "Usage: %s [-a CPU -p PRIO -i USEC -r USEC -s NS -l N -v]\n\n"
      "       -h                 Show help\n"
      "       -a CPU             Pin the real time thread to CPU\n"
      "       -p PRIO            RT thread priority\n"
      "       -i USEC            RT thread wake-up interval (default: 10ms)\n"
      "       -r USEC            non-RT main thread refresh interval\n"
      "       -l N_CYCLES        RT thread cycles amount (default: infinite)\n"
      "       -s NS              Common start time reference\n"
      "       -g                 Use GPIO instead of serial\n"
      "       -P USEC            Do polling to wakeup with specified margin\n"
      "       -v                 Verbose\n"
      "\n",
      argv[0]);
}

static void poll_wakeup(struct timespec ts, int margin) {
  int ret;
  struct timespec ts_prev, current;

  ts_prev = ts;
  substract_ns(&ts_prev, margin * 1000);

  ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &ts_prev, NULL);
  if (ret) {
    fprintf(stderr, "clock_nanosleep returned error: %d, aborting...\n", ret);
    exit(EXIT_FAILURE);
  }

  do {
    clock_gettime(CLOCK_REALTIME, &current);
  } while(calcdiff_ns_signed(ts, current) > 1000);
}

static void *print_thread(void *p) {
  (void)p;

  for (;;) {
    if (thread_params.max_cycles &&
        nb_cycles >= ((unsigned int)thread_params.max_cycles))
      break;

    printf("Interval: %10" PRIu64 " Target: %10" PRIu64 "\n", thread_params.interval / 1000, main_params.target_interval);
    printf("\033[%dA", 1);

    usleep(100000);
  }

  pthread_exit(NULL);
}

/*
 * Real-time thread:
 */
static void *pulse_thread(void *p) {
  (void)p;
  struct timespec next;
  int ret;
  cpu_set_t mask;

  uint64_t next_increment = 0;
  uint64_t end_t = 0;
  uint64_t cur_t = 0;
  uint64_t i_t = 0;
  uint64_t i_s = thread_params.interval / 1000;
  uint64_t i_c = thread_params.interval / 1000;

  // Set thread CPU affinity
  if (thread_params.affinity_cpu != -1) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  clock_gettime(CLOCK_REALTIME, &next);

  if (thread_params.start_ts) {
    if (thread_params.start_ts < ts_to_uint(next)) {
      fprintf(stderr, "start timestamp is in the past, aborting...\n");
      exit(EXIT_FAILURE);
    }
    if (thread_params.start_ts > (ts_to_uint(next) + UINT64_C(3600000000000))) {
      fprintf(stderr, "start timestamp is too high, aborting...\n");
      exit(EXIT_FAILURE);
    }
    next = uint_to_ts(thread_params.start_ts);
  }

  // Pulse loop
  for (nb_cycles = 0;; nb_cycles++) {
    if (thread_params.max_cycles &&
        nb_cycles >= ((unsigned int)thread_params.max_cycles))
      break;

    if (use_gpio)
      toggle_gpio();
    else
      send_pulse();

    add_ns(&next, thread_params.interval);

    if(main_params.interval_input && (thread_params.interval != (main_params.target_interval * 1000))) {

      i_c = thread_params.interval / 1000;

      // If target interval changed
      if(i_t != main_params.target_interval) {

        i_t = main_params.target_interval;
        i_s = i_c;
        cur_t = 0;
        next_increment = 0;

        if(i_t < i_s)
          end_t = (main_params.transition_time * USEC_PER_SEC * USEC_PER_SEC * (i_c - i_t)) / (MOTOR_STEPS * i_t * i_c);
        else
          end_t = (main_params.transition_time * USEC_PER_SEC * USEC_PER_SEC * (i_t - i_c)) / (MOTOR_STEPS * i_t * i_c);

      }

      if(next_increment) {
        if(cur_t > next_increment) {
          i_c += i_t < i_s ? -1 : 1;
          next_increment = 0;
        }
      } else {
        uint64_t next_i = i_t < i_s ? i_c - 1 : i_c + 1;
        // Compute time at which we will need to increment / decrement the interval

        if(i_t < i_s)
          next_increment = (end_t * i_t * (i_s - next_i)) / ((i_s - i_t) * next_i);
        else
          next_increment = (end_t * i_t * (next_i - i_s)) / ((i_t - i_s) * next_i);

        // If next increment time is before next interval
        if(next_increment < cur_t + i_c) {
          if(i_t < i_s) {
            i_c = (i_t * i_s * end_t) / ((i_s - i_t) * (cur_t + i_c) + end_t * i_t);
            i_c = i_c < i_t ? i_t : i_c;
          }
          else {
            i_c = (i_t * i_s * end_t) / (end_t * i_t - (i_t - i_s) * (cur_t + i_c));
            i_c = i_c > i_t ? i_t : i_c;
          }
          next_increment = 0;
        }
      }

      cur_t += i_c;
      if(i_c < 50) {
        fprintf(stderr, "Interval too small, exiting..\n");
        exit(EXIT_FAILURE);
      }
      if(i_c > USEC_PER_SEC) {
        fprintf(stderr, "Interval too big, exiting..\n");
        exit(EXIT_FAILURE);
      }
      thread_params.interval = i_c * 1000;
    }

    if(thread_params.poll) {
      poll_wakeup(next, thread_params.poll_margin);
    }
    else {
      ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &next, NULL);
      if (ret) {
        fprintf(stderr, "clock_nanosleep returned error: %d, aborting...\n", ret);
        exit(EXIT_FAILURE);
      }
    }

  }

  pthread_exit(NULL);
}

/*
 * Main thread:
 *  - Has non-real time priority
 *  - Handles the IO and creates the real time thread
 */
int main(int argc, char *argv[]) {
  pthread_t thread, print_pthread;
  struct sched_param param;
  pthread_attr_t attr;

  // Default configuration values
  thread_params.interval = 100000 * 1000;
  thread_params.max_cycles = 0;
  thread_params.priority = 98;
  thread_params.affinity_cpu = -1;
  thread_params.start_ts = 0;
  thread_params.poll = 0;
  thread_params.poll_margin = 75;
  main_params.refresh_rate = 50000;
  main_params.verbose = 0;
  main_params.interval_input = 0;
  use_gpio = 0;

  // Lock all current and future pages from preventing of being paged to
  // swap
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    fprintf(stderr, "mlockall failed");
    exit(EXIT_FAILURE);
  }

  // Process bash options
  process_options(argc, argv);

  main_params.target_interval = thread_params.interval / 1000;

  if (use_gpio)
    enable_gpio(86);
  else
    enable_pulse();

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&attr)) {
    fprintf(stderr, "init pthread attributes failed\n");
    exit(EXIT_FAILURE);
  }
  // Set a specific stack size
  if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN)) {
    fprintf(stderr, "pthread setstacksize failed\n");
    exit(EXIT_FAILURE);
  }
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
    fprintf(stderr, "pthread setschedpolicy failed\n");
    exit(EXIT_FAILURE);
  }
  param.sched_priority = thread_params.priority;
  if (pthread_attr_setschedparam(&attr, &param)) {
    fprintf(stderr, "pthread setschedparam failed\n");
    exit(EXIT_FAILURE);
  }
  /* Use scheduling parameters of attr */
  if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
    fprintf(stderr, "pthread setinheritsched failed\n");
    exit(EXIT_FAILURE);
  }

  // Create the real time thread
  if (pthread_create(&thread, &attr, pulse_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create pulse thread");

  // Create the print thread
  if (pthread_create(&print_pthread, NULL, print_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create print thread");

  // Verbose loop
  for (;;) {

    if (main_params.interval_input) {
      uint64_t user_input;
      scanf("%" PRIu64, &user_input);
      if(user_input)
        main_params.target_interval = user_input;
    }
    else if (main_params.verbose) {
      usleep(main_params.refresh_rate);
      printf("Toggles: %9" PRIu64 "\n", nb_cycles);
      printf("\033[%dA", 1);
    }

    if (thread_params.max_cycles)
      if (thread_params.max_cycles == nb_cycles) break;
  }

  exit(EXIT_SUCCESS);
}

/* Process bash options
 */
static void process_options(int argc, char *argv[]) {
  for (;;) {
    int c = getopt(argc, argv, "a:hgi:l:p:r:s:vP:U:");

    if (c == -1) break;

    switch (c) {
      case 'a':
        thread_params.affinity_cpu = atoi(optarg);
        break;
      case 'h':
        help(argv);
        exit(EXIT_SUCCESS);
        break;
      case 'g':
        use_gpio = 1;
        break;
      case 'i':
        thread_params.interval = strtoull(optarg, NULL, 10) * 1000;
        break;
      case 'l':
        thread_params.max_cycles = atoi(optarg);
        break;
      case 'p':
        thread_params.priority = atoi(optarg);
        break;
      case 's':
        thread_params.start_ts = strtoull(optarg, NULL, 10);
        break;
      case 'r':
        main_params.refresh_rate = atoi(optarg);
        break;
      case 'v':
        main_params.verbose = 1;
        break;
      case 'P':
        thread_params.poll = 1;
        thread_params.poll_margin = atoi(optarg);
        break;
      case 'U':
        main_params.interval_input = 1;
        main_params.transition_time = atoi(optarg);
        break;
    }
  }
}
